from dbterd.adapters.targets.dbml.engine.meta import Column, Ref, Table


def parse(manifest, catalog, **kwargs):
    # Parse Table
    tables = get_tables(manifest, catalog)
    # -- apply selection
    select_rule = (kwargs.get("select") or "").lower().split(":")
    if select_rule[0].startswith("schema"):
        select_rule = select_rule[-1]
        tables = [
            x
            for x in tables
            if x.schema.startswith(select_rule)  # --select schema:analytics
            or f"{x.database}.{x.schema}".startswith(
                select_rule
            )  # --select schema:db.analytics
        ]
    else:
        select_rule = select_rule[-1]  # only take care of name
        tables = [x for x in tables if x.name.startswith(select_rule)]

    # -- apply exclusion (take care of name only)
    tables = [
        x
        for x in tables
        if kwargs.get("exclude") is None or not x.name.startswith(kwargs.get("exclude"))
    ]

    # Parse Ref
    relationships = get_relationships(manifest)
    table_names = [x.name for x in tables]
    relationships = [
        x
        for x in relationships
        if x.table_map[0] in table_names and x.table_map[1] in table_names
    ]

    # Fullfill columns in Tables (due to `select *`)
    for relationship in relationships:
        for table in tables:
            table_columns = [x.name for x in table.columns]
            if (
                table.name == relationship.table_map[0]
                and relationship.column_map[0] not in table_columns
            ):
                table.columns.append(Column(name=relationship.column_map[0]))
            if (
                table.name == relationship.table_map[1]
                and relationship.column_map[1] not in table_columns
            ):
                table.columns.append(Column(name=relationship.column_map[1]))

    # Build DBML content
    dbml = "//Tables (based on the selection criteria)\n"
    for table in tables:
        dbml += f"//--configured at schema: {table.database}.{table.schema}\n"
        dbml += """Table \"{table}\"{{\n{columns}\n}}\n""".format(
            table=table.name,
            columns="\n".join([f'    "{x.name}" {x.data_type}' for x in table.columns]),
        )

    dbml += "//Refs (based on the DBT Relationship Tests)\n"
    for rel in relationships:
        key_from = f'"{rel.table_map[1]}"."{rel.column_map[1]}"'
        key_to = f'"{rel.table_map[0]}"."{rel.column_map[0]}"'
        dbml += f"Ref: {key_from} > {key_to}\n"

    return dbml


def get_tables(manifest, catalog):

    tables = [
        Table(
            name=x,
            raw_sql=get_compiled_sql(manifest.nodes[x]),
            database=manifest.nodes[x].database.lower(),
            schema=manifest.nodes[x].schema_.lower(),
            columns=[],
        )
        for x in manifest.nodes
        if x.startswith("model")
    ]

    for table in tables:
        # Pull columns from the catalog and use the data types declared there
        # Catalog is our primary source of information about the target db
        cat_columns = catalog.nodes[table.name].columns
        for column, metadata in cat_columns.items():
            table.columns.append(Column(name=column, data_type=metadata.type))
        # Handle cases where columns don't exist yet, but are in manifest
        man_columns = manifest.nodes[table.name].columns
        for column in man_columns:
            if column in cat_columns:
                # Already exists in the remote
                continue
            table.columns.append(Column(name=column, data_type="UNKNOWN"))

    return tables


def get_relationships(manifest):
    refs = [
        Ref(
            name=x,
            table_map=manifest.parent_map[x],
            column_map=[
                manifest.nodes[x].test_metadata.kwargs.get("field", "unknown"),
                manifest.nodes[x].test_metadata.kwargs.get("column_name", "unknown"),
            ],
        )
        for x in manifest.nodes
        if (
            x.startswith("test")
            and "relationship" in x.lower()
            and manifest.nodes[x].meta.get("ignore_in_erd", "0") == "0"
        )
    ]

    # remove duplicates
    if refs:
        distinct_list = [refs[0]]
        for ref in refs:
            distinct_maps = [str((x.table_map, x.column_map)) for x in distinct_list]
            if str((ref.table_map, ref.column_map)) not in distinct_maps:
                distinct_list.append(ref)

        return distinct_list

    return []


def get_compiled_sql(manifest_node):
    if hasattr(manifest_node, "compiled_sql"):  # up to v6
        return manifest_node.compiled_sql

    if hasattr(manifest_node, "compiled_code"):  # from v7
        return manifest_node.compiled_code

    if hasattr(
        manifest_node, "columns"
    ):  # nodes having no compiled but just list of columns
        return """select
            {columns}
        from {table}
        """.format(
            columns="\n".join(
                [
                    f"{x} as {manifest_node.columns[x].data_type or 'varchar'},"
                    for x in manifest_node.columns
                ]
            ),
            table=f"{manifest_node.database}.{manifest_node.schema}.undefined",
        )

    return manifest_node.raw_sql  # fallback to raw dbt code



from dataclasses import dataclass
from typing import List, Optional, Tuple


@dataclass
class Column:
    """
    Sample DBML: id varchar [primary key]
    """

    name: str
    data_type: str = "varchar"


@dataclass
class Table:
    """
    Sample DBML:
    Table posts {
        id varchar [primary key]
    }
    """

    name: str
    database: str
    schema: str
    columns: Optional[List[Column]] = None
    raw_sql: Optional[str] = None


@dataclass
class Ref:
    """Sample DBML: Ref: posts.user_id > users.id"""

    name: str
    table_map: Tuple[str, str]
    column_map: Tuple[str, str]



from importlib import import_module


def load_executor(name: str):
    try:
        return import_module(name=f".{name}", package="dbterd.adapters.targets")
    except ModuleNotFoundError as exc:
        if exc.name == "dbterd.adapters.targets." + name:
            raise Exception(f"Could not find adapter target type {name}!")
        raise



import os
import json
import sys

from dbt_artifacts_parser import parser

if sys.platform == "win32":
    from ctypes import WinDLL, c_bool
else:
    WinDLL = None
    c_bool = None


def load_file_contents(path: str, strip: bool = True) -> str:
    path = convert_path(path)
    with open(path, "rb") as handle:
        to_return = handle.read().decode("utf-8")

    if strip:
        to_return = to_return.strip()

    return to_return


def open_json(fp):
    """Json loading utility, leveraging long path fixes"""
    return json.loads(load_file_contents(fp))


def convert_path(path: str) -> str:
    """Convert a path which might be >260 characters long, to one
    that will be writable/readable on Windows.
    On other platforms, this is a no-op.
    """
    # some parts of python seem to append '\*.*' to strings, better safe than
    # sorry.
    if len(path) < 250:
        return path
    if _supports_long_paths():
        return path

    prefix = "\\\\?\\"
    # Nothing to do
    if path.startswith(prefix):
        return path

    path = _win_prepare_path(path)

    # add the prefix. The check is just in case os.getcwd() does something
    # unexpected - I believe this if-state should always be True though!
    if not path.startswith(prefix):
        path = prefix + path
    return path


def _supports_long_paths() -> bool:
    if sys.platform != "win32":
        return True
    # Eryk Sun says to use `WinDLL('ntdll')` instead of `windll.ntdll` because
    # of pointer caching in a comment here:
    # https://stackoverflow.com/a/35097999/11262881
    # I don't know exaclty what he means, but I am inclined to believe him as
    # he's pretty active on Python windows bugs!
    else:
        try:
            dll = WinDLL("ntdll")
        except OSError:  # I don't think this happens? you need ntdll to run python
            return False
        # not all windows versions have it at all
        if not hasattr(dll, "RtlAreLongPathsEnabled"):
            return False
        # tell windows we want to get back a single unsigned byte (a bool).
        dll.RtlAreLongPathsEnabled.restype = c_bool
        return dll.RtlAreLongPathsEnabled()


def _win_prepare_path(path: str) -> str:
    """Given a windows path, prepare it for use by making sure it is absolute
    and normalized.
    """
    path = os.path.normpath(path)

    # if a path starts with '\', splitdrive() on it will return '' for the
    # drive, but the prefix requires a drive letter. So let's add the drive
    # letter back in.
    # Unless it starts with '\\'. In that case, the path is a UNC mount point
    # and splitdrive will be fine.
    if not path.startswith("\\\\") and path.startswith("\\"):
        curdrive = os.path.splitdrive(os.getcwd())[0]
        path = curdrive + path

    # now our path is either an absolute UNC path or relative to the current
    # directory. If it's relative, we need to make it absolute or the prefix
    # won't work. `ntpath.abspath` allegedly doesn't always play nice with long
    # paths, so do this instead.
    if not os.path.splitdrive(path)[0]:
        path = os.path.join(os.getcwd(), path)

    return path


def read_manifest(manifest_path: str, manifest_version: int):
    """Reads in the manifest file, with optional version specification"""
    manifest_dict = open_json(f"{manifest_path}/manifest.json")
    parser_version = (
        f"parse_manifest_v{manifest_version}" if manifest_version else "parse_manifest"
    )
    parse_func = getattr(parser, parser_version)
    manifest_obj = parse_func(manifest=manifest_dict)
    return manifest_obj


def read_catalog(catalog_path):
    """reads and parses the catalog file"""
    catalog_dict = open_json(f"{catalog_path}/catalog.json")
    return parser.parse_catalog(catalog=catalog_dict)



import importlib.metadata

import click

from dbterd.adapters.worker import DbtWorker
from dbterd.cli import params
from dbterd.helpers import jsonify
from dbterd.helpers.log import logger

__version__ = importlib.metadata.version("dbterd")


# dbterd
@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
    no_args_is_help=True,
    epilog="Specify one of these sub-commands and you can find more help from there.",
)
@click.version_option(__version__)
@click.pass_context
def dbterd(ctx, **kwargs):
    """Tools for producing diagram-as-code"""


# dbterd debug
@dbterd.command(name="debug")
@click.pass_context
@params.common_params
def debug(ctx, **kwargs):
    """Inspect the hidden magics"""
    logger.info("**Arguments used**")
    logger.debug(jsonify.to_json(kwargs))
    logger.info("**Arguments used**")
    logger.debug(jsonify.to_json(ctx.obj))


# dbterd run
@dbterd.command(name="run")
@click.pass_context
@params.common_params
def run(ctx, **kwargs):
    """Run the convert"""
    DbtWorker(ctx).run(**kwargs)



import abc
from asyncio.log import logger

import click

from dbterd.adapters import factory
from dbterd.helpers import cli_messaging
from dbterd.helpers import file as file_handlers


class Executor(abc.ABC):
    ctx: click.Context

    def __init__(self, ctx) -> None:
        super().__init__()
        self.ctx = ctx
        self.filename_manifest = "manifest.json"
        self.filename_catalog = "catalog.json"

    @abc.abstractmethod
    def run(self, **kwargs):
        self.__run_by_strategy(**kwargs)

    def __read_manifest(self, mp: str, mv: int = None):
        cli_messaging.check_existence(mp, self.filename_manifest)
        conditional = f" or provided version {mv} is incorrect" if mv else ""
        with cli_messaging.handle_read_errors(self.filename_manifest, conditional):
            return file_handlers.read_manifest(mp, mv)

    def __read_catalog(self, cp: str):
        cli_messaging.check_existence(cp, self.filename_catalog)
        with cli_messaging.handle_read_errors(self.filename_catalog):
            return file_handlers.read_catalog(cp)

    def __run_by_strategy(self, **kwargs):
        target_module = factory.load_executor(name=kwargs["target"])
        operation_dispatcher = getattr(target_module, "run_operation_dispatcher")
        strategy_func = operation_dispatcher.get(
            f"{kwargs['target']}_{kwargs['algo']}",
            getattr(target_module, "operation_default"),
        )
        manifest = self.__read_manifest(
            mp=kwargs["manifest_path"], mv=kwargs["manifest_version"]
        )
        catalog = self.__read_catalog(
            # TODO: Find a cleaner way to pass these args
            cp=kwargs.get("catalog_path", kwargs["manifest_path"])
        )

        result = strategy_func(manifest=manifest, catalog=catalog, **kwargs)
        path = kwargs["output"] + f"/{result[0]}"
        with open(path, "w") as f:
            logger.info(path)
            f.write(result[1])



from pathlib import Path


def default_manifest_path() -> str:
    paths = list(Path.cwd().parents)
    paths.insert(0, Path.cwd())
    return str(
        next((x for x in paths if (x / "target/manifest.json").exists()), Path.cwd())
        / "target"
    )


def default_output_path() -> str:
    return str(Path.cwd() / "target")


def default_target() -> str:
    return "dbml"


def deafult_algo() -> str:
    return "test_relationship"



class ObjectView(object):
    """
    Convert nested dict to dynamic object
    """

    def __init__(self, d, nested=True):
        self.origin = d
        for a, b in d.items():
            if nested:
                if isinstance(b, (list, tuple)):
                    setattr(
                        self,
                        a,
                        [ObjectView(x) if isinstance(x, dict) else x for x in b],
                    )
                else:
                    setattr(self, a, ObjectView(b) if isinstance(b, dict) else b)
            else:
                setattr(self, a, b)

    def has_field(self, field):
        fields = field.split(".")
        if len(fields) == 0:
            return False

        obj = self.origin
        for f in fields:
            if f not in obj:
                return False
            obj = obj[f]

        return True



import functools

import click

from dbterd import default


def common_params(func):
    @click.option(
        "--manifest-path",
        "-mp",
        help="Specified the full path to dbt manifest.json file",
        default=default.default_manifest_path(),
        type=click.STRING,
    )
    @click.option(
        "--manifest-version",
        "-mv",
        help="Specified dbt manifest.json version",
        default=None,
        type=click.STRING,
    )
    @click.option(
        "--target",
        "-t",
        help="Target to the diagram-as-code platform",
        default=default.default_target(),
        type=click.STRING,
    )
    @click.option(
        "--output",
        "-o",
        help="Output the result file. Default to the same target dir",
        default=default.default_output_path(),
        type=click.STRING,
    )
    @click.option(
        "--algo",
        "-a",
        help="Specified algorithm in the way to detect diagram connectors",
        default=default.deafult_algo(),
        type=click.STRING,
    )
    @click.option(
        "--select",
        "-s",
        help="Selecttion criteria. Support 'starts with' a string",
        default=None,
        type=click.STRING,
    )
    @click.option(
        "--exclude",
        "-ns",
        help="Exclusion criteria. Support 'not starts with' a string",
        default=None,
        type=click.STRING,
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



from dbterd.adapters.targets.dbml.constants import Strategies
from dbterd.adapters.targets.dbml.strategies import dbml_test_relationship, default

operation_default = default.default
run_operation_dispatcher = {
    Strategies.DBML_TEST_RELATIONSHIP: dbml_test_relationship.run,
}



import contextlib
import json
from pathlib import Path

import click


@contextlib.contextmanager
def handle_read_errors(filename, conditional_msg: str = ""):
    try:
        yield
    except (json.JSONDecodeError, ValueError):
        raise click.FileError(
            filename,
            f"File {filename} is corrupted{conditional_msg}, please rebuild",
        )


def check_existence(path_str: str, filename: str) -> None:
    path = Path(path_str)
    if not path.is_dir():
        raise click.FileError(filename, f"Path {path_str} does not exist")
    elif not (path / filename).is_file():
        raise click.FileError(
            filename, f"File {filename} does not exist in directory {path_str}"
        )



import dataclasses
import json


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if type(o).__name__ == "module":
            return o.__name__
        return super().default(o)


def __mask(obj: str, masks: list = []):
    obj_dict = json.loads(obj)
    for key, value in obj_dict.items():
        if key in masks:
            obj_dict[key] = value[0:5] + "***********"
        if isinstance(value, dict):
            obj_dict[key] = __mask(json.dumps(value, cls=EnhancedJSONEncoder), masks)

    return obj_dict


def to_json(obj, masks=[]):
    if not obj:
        return {}
    mask_dict = obj
    if "__dict__" in mask_dict:
        mask_dict = __mask(json.dumps(obj.__dict__, cls=EnhancedJSONEncoder), masks)
    return json.dumps(mask_dict, indent=4, cls=EnhancedJSONEncoder)
